#pragma once
#include<cmath>
#include<cstdlib>
#include<vector>
using namespace std;
#define PI 3.14159265

int int_round(float x){
	if( x >= 0.0)
		return (int)(x + 0.5f);
	else
		return (int)(x - 0.5f);
}

vector<float> GenerateGaussian1D(float variance){
	int n = int_round(2.6f*variance);
	n = max(1, n);
	vector<float> weights1d;
	weights1d.resize(2*n+1);
	
	weights1d[n] = 1.0;

	for(int i = 1; i <= n; i++){
		weights1d[n+i] = weights1d[n-i] = (float)exp(-0.5*pow((double)i/variance, 2));
	}
	return weights1d;
}

vector<float> GenerateHalfGaussian1D(float variance){
	int n = int_round(2.6f*variance);
	n = max(1, n);
	vector<float> weights1d;
	weights1d.resize(n+1);

	weights1d[0] = 1.0;
	for(int i = 1; i <= n; i++){
		weights1d[i] = (float)exp(-0.5*pow((double)i/variance, 2));
	}
	return weights1d;
}

floatimage GenerateGaussian2D(float variance){
	vector<float> gaussian1d = GenerateGaussian1D(variance);
	floatimage gaussian2d((int)gaussian1d.size(), (int)gaussian1d.size());
	double total = 0.0;
	for(int i = 0; i < (int)gaussian1d.size(); i++){
		for(int j = 0; j < (int)gaussian1d.size(); j++){
			gaussian2d.line[i][j].r = gaussian2d.line[i][j].g = gaussian2d.line[i][j].b = gaussian1d[i]*gaussian1d[j];
			total += gaussian2d.line[i][j].r;
		}
	}

	for(int i = 0; i < (int)gaussian1d.size(); i++){
		for(int j = 0; j < (int)gaussian1d.size(); j++){
			gaussian2d.line[i][j].r /= (float)total;
			gaussian2d.line[i][j].g /= (float)total;
			gaussian2d.line[i][j].b /= (float)total;
		}
	}
	return gaussian2d;
}

inline float Gaussian(float x, float var){
	return (float)exp(-0.5f*pow(x/var, (int)2));
}


class gaussianTable{
public:
	gaussianTable(float var, int qsize);

	vector<float> table;
	float operator()(float x);
	float qstep;
};


gaussianTable::gaussianTable(float var, int qsize){
	table.resize(qsize);
	table[0] = 1.0;
	float range = var*2.6f;
	qstep = range/qsize;
	float x = 0.0;

	for(int i = 0; i < qsize; i++){
		table[i] = (float)exp(-0.5*pow((double)x/var, 2));
		x += qstep;
	}
}

float gaussianTable::operator ()(float x){
	int i = (float)(abs(x)/qstep);
	if(i < (int)table.size())
		return table[i];
	else
		return 0.0f;
}