#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <fstream>
using namespace std;

#include "floatimage.h"
#include "gaussian.h"

ofstream logfile;


void printHelp()
{
	fprintf(stderr, "Bilateral Filtering.\n");
	fprintf(stderr, "HDRSHOPFLAGS: \n\n");
	fprintf(stderr, "HDRSHOPVERSION: 1.0.1\n\n");
	fprintf(stderr, "USAGE: bilateral filtering [input.pfm] [output.pfm] [options]\n\n");
	fprintf(stderr, "OPTIONS:\n");
	fprintf(stderr, "  /D=[float=5.0]		sigma d\n");
	fprintf(stderr, "  /R=[float=0.4]       sigma r\n");
	fprintf(stderr, "  /A=[float=0.09]       a\n");
	fprintf(stderr, "  /L=[float=3.0]      Lwhite\n");
	
	
}


void printError(int err)
{
	switch (err)
	{
		case 0:
			fprintf(stderr, "No error\n");
			break;
		case 1:
			fprintf(stderr, "ERROR: No input file specified\n");
			break;
		case 2:
			fprintf(stderr, "ERROR: No output file specified\n");
			break;
		case 3:
			fprintf(stderr, "ERROR: Cannot open input file\n");
			break;
		case 4:
			fprintf(stderr, "ERROR: Cannot open output file\n");
			break;
	}
}




void doError(int err)
{
	printError(err);
	exit(err);
}



//bilateral Filtering On R(Y)
floatimage BilateralFiteringY(float vd, float vr, floatimage & in){
	floatimage tmp;
	tmp.Allocate(in.sizex, in.sizey);
	floatimage kernel = GenerateGaussian2D(vd);
	gaussianTable g(vr, 1500);
	//vector<float> halfGaussian = GenerateHalfGaussian1D(vr);


	for (int y = 0; y < in.sizey; y++) 
	{
		fprintf(stderr,  "bi filtering line: %d\n", y);
		for (int x = 0; x < in.sizex; x++) 
		{
			float	r = 0;
			float	kr = 0;

			for (int ky = 0; ky < kernel.sizey; ky++) 
			{
				for (int kx = 0; kx < kernel.sizex; kx++)
				{
					int	cx = x + kx - kernel.sizex / 2;
					int	cy = y + ky - kernel.sizey / 2;

					if (!((cx < 0) || (cy < 0) || (cx >= in.sizex) || (cy >= in.sizey))) 
					{
						float w = g((float)in.line[cy][cx].r - in.line[y][x].r);
						r += w*kernel.line[ky][kx].r * in.line[cy][cx].r;
						kr += w*kernel.line[ky][kx].r;
					}
				}
			}

			tmp.line[y][x].r = r / kr;
			tmp.line[y][x].g = 0;
			tmp.line[y][x].b = 0;
		}
	}
	return tmp;
}


floatimage getIntensity(const floatimage & in){
	floatimage tmp;
	tmp.Allocate(in.sizex, in.sizey);
	for(int y = 0; y < in.sizey; y++){
		for(int x = 0; x < in.sizex; x++){
			//float I = max(in.line[y][x].r, in.line[y][x].g);
			//I = max(in.line[y][x].b, I);
			//tmp.line[y][x].r = I;

			tmp.line[y][x].r =   0.2989f*in.line[y][x].r + 0.5866f*in.line[y][x].g + 0.1145f*in.line[y][x].b; //Y
			//tmp.line[y][x].r =  (1/61.0f)*(20*in.line[y][x].r + 40*in.line[y][x].g + in.line[y][x].b); //Y
			tmp.line[y][x].g = 0.0f;
			tmp.line[y][x].b = 0.0f;
		}
	}
	return tmp;
}

floatimage getColor(const floatimage & in, const floatimage & I){
	floatimage tmp;
	tmp.Allocate(in.sizex, in.sizey);
	for(int y = 0; y < in.sizey; y++){
		for(int x = 0; x < in.sizex; x++){
			//float rgb = in.line[y][x].r + in.line[y][x].g + in.line[y][x].b;
			tmp.line[y][x].r =  in.line[y][x].r/I.line[y][x].r;
			tmp.line[y][x].g =  in.line[y][x].g/I.line[y][x].r;
			tmp.line[y][x].b =  in.line[y][x].b/I.line[y][x].r;
		}
	}
	return tmp;
}

floatimage getLog(const floatimage & in){
	floatimage tmp;
	tmp.Allocate(in.sizex, in.sizey);
	for(int y = 0; y < in.sizey; y++){
		for(int x = 0; x < in.sizex; x++){
			if(in.line[y][x].r == 0.0f)
				tmp.line[y][x].r = - 7.0;
			else
				tmp.line[y][x].r = (float)log10((double)in.line[y][x].r);
			tmp.line[y][x].g = 0;
			tmp.line[y][x].b = 0;
		}
	}
	return tmp;
}

floatimage getFromLog(const floatimage & in){
	floatimage tmp;
	tmp.Allocate(in.sizex, in.sizey);
	for(int y = 0; y < in.sizey; y++){
		for(int x = 0; x < in.sizex; x++){
			tmp.line[y][x].r = (float)pow(10.0 ,(double)in.line[y][x].r );
			tmp.line[y][x].g = 0;
			tmp.line[y][x].b = 0;
		}
	}
	return tmp;
}

floatimage getImageFromColorIntensity(const floatimage & color, const floatimage & intensity){
	floatimage tmp;
	tmp.Allocate(color.sizex, color.sizey);
	for(int y = 0; y < color.sizey; y++){
		for(int x = 0; x < color.sizex; x++){
			//float rgb = in.line[y][x].r + in.line[y][x].g + in.line[y][x].b;
			tmp.line[y][x].r =  color.line[y][x].r*intensity.line[y][x].r;
			tmp.line[y][x].g =  color.line[y][x].g*intensity.line[y][x].r;
			tmp.line[y][x].b =  color.line[y][x].b*intensity.line[y][x].r;
		}
	}
	return tmp;
}

floatimage ConvertRgbToYuv(const floatimage & in){
	floatimage tmp(in.sizex, in.sizey);
	
	for(int y = 0; y < in.sizey; y++){
		for(int x = 0; x < in.sizex; x++){
			tmp.line[y][x].r =  0.2989f*in.line[y][x].r + 0.5866f*in.line[y][x].g + 0.1145f*in.line[y][x].b; //Y
			tmp.line[y][x].g = -0.1687f*in.line[y][x].r - 0.3312f*in.line[y][x].g + 0.5000f*in.line[y][x].b; //U
			tmp.line[y][x].b =  0.5000f*in.line[y][x].r - 0.4183f*in.line[y][x].g - 0.0816f*in.line[y][x].b; //V
		}
	}
	return tmp;
}

floatimage ConvertYuvToRgb(const floatimage & in){
	floatimage tmp(in.sizex, in.sizey);
	
	for(int y = 0; y < in.sizey; y++){
		
		for(int x = 0; x < in.sizex; x++){
			tmp.line[y][x].r = in.line[y][x].r + 1.4022f*in.line[y][x].b; //Y
			tmp.line[y][x].g = in.line[y][x].r - 0.3456f*in.line[y][x].g - 0.7145f*in.line[y][x].b; //U
			tmp.line[y][x].b = in.line[y][x].r + 1.7710f*in.line[y][x].g;
		}
	}
	return tmp;
}

inline float climp(float x){
	if( x < 0.0f)
		return 0.0f;
	else if( x > 1.0f)
		return 1.0f;
	else
		return x;
}


int main(int argc, char *argv[])
{
	logfile.open("logfile.txt", ios::app);
	logfile << "------- arg --------" << endl;
	for(int i = 0; i < argc; i++)
		logfile << argv[i] << endl;

	char *inputfilename = 0;
	char *outputfilename = 0;

	float sigma_d = 5.0f;
	float sigma_r = 0.4f;
	float a = 0.09f;
	float Lwhite = 3.0f;
	

	if (argc == 1)								// if no arguments specified - print help and exit
	{
		printHelp();
		exit(0);
	}

	for (int x = 1; x < argc; x++)				// parse the command line
	{
		if (argv[x][0] == '-')
		{
			switch(argv[x][1]) {
			case '-':
				if (strncmp("help", &argv[x][2], 4) == 0)
				{
					printHelp();
					exit(0);
				}
				break;
			case 'd':
			case 'D':
				sigma_d = (float)atof(argv[x] + 3);
				break;				
			case 'r':
			case 'R':
				sigma_r = (float)atof(argv[x] + 3);
				break;
			case 'a':
			case 'A':
				a = (float)atof(argv[x] + 3);
				break;
			case 'l':
			case 'L':
				Lwhite = (float)atof(argv[x] + 3);
				break;
			}
		}
		else if (argv[x][0] == '/')
		{
			switch(argv[x][1]) {
			case '?':
				printHelp();
				exit(0);
				break;
			case 'D':
				sigma_d = (float)atof(argv[x] + 3);
				break;
			case 'R':
				sigma_r = (float)atof(argv[x] + 3);
				break;
			case 'A':
				a = (float)atof(argv[x] + 3);
				break;
			case 'L':
				Lwhite = (float)atof(argv[x] + 3);
				break;
			}
		}
		else
		{
			if (inputfilename == 0) 
				inputfilename = argv[x];
			else
				outputfilename = argv[x];
		}
	}


	logfile << "----- start ------ " << endl;

	if (!inputfilename)					// make sure we have specified an input and an output file
		doError(1);
	if (!outputfilename) 
		doError(2);

	floatimage fimg;					// load up the input file
	if (!fimg.LoadPFM(inputfilename))
		doError(3);
	fprintf(stderr, "\n\n");
	fprintf(stderr, "input file: %s\n", inputfilename);
	fprintf(stderr, "output file: %s\n", outputfilename);

	//sigma_d = 0.02f*max(fimg.sizex, fimg.sizey);

	floatimage intensity = getIntensity(fimg);
	floatimage color = getColor(fimg, intensity);
	floatimage detail = getLog(intensity);

	floatimage base = BilateralFiteringY(sigma_d, sigma_r, detail);
	detail -= base;

	// contrast reduction !!!

	float Lw = 0.0f;
	for(int y = 0;  y < base.sizey; y++){
		for(int x = 0; x < base.sizex; x++){
			Lw += base.line[y][x].r;
		}
	}

	Lw /= base.sizex*base.sizey; //average
	Lw = (float)pow(10.0, (double)Lw);

	base = getFromLog(base);

	for(int y = 0;  y < base.sizey; y++){
		for(int x = 0; x < base.sizex; x++){
			base.line[y][x].r = (base.line[y][x].r/Lw)*a;
			base.line[y][x].r = base.line[y][x].r/(1.0f + base.line[y][x].r);
			base.line[y][x].r = base.line[y][x].r*(1+base.line[y][x].r/(Lwhite*Lwhite))/(1+base.line[y][x].r);
		}
	}

	ConvertYuvToRgb(base).SavePFM("base.pfm");
	detail = getFromLog(detail);
	for(int y = 0;  y < base.sizey; y++){
		for(int x = 0; x < base.sizex; x++){
			intensity.line[y][x].r = base.line[y][x].r*detail.line[y][x].r;
		}
	}

	ConvertYuvToRgb(detail).SavePFM("detail.pfm");

	ConvertYuvToRgb(intensity).SavePFM("intensity_nocolor.pfm");

	fimg = getImageFromColorIntensity(color, intensity);

	for(int y = 0; y < fimg.sizey; y++){
		for(int x = 0; x < fimg.sizex; x++){
			fimg.line[y][x].r = climp((float)pow((double)fimg.line[y][x].r,1.0));
			fimg.line[y][x].g = climp((float)pow((double)fimg.line[y][x].g,1.0));
			fimg.line[y][x].b = climp((float)pow((double)fimg.line[y][x].b,1.0));
		}
	}

	if(!fimg.SavePFM(outputfilename))	// save the output file
		doError(4);
	
	logfile << "-------- end prog --------" << endl;
	logfile.close();
	fprintf(stderr, "Success!\n");		// display message to the user

	return 0;
}